use super::*;

/// Public keys for a given DID kind in the Morpheus subtree in a vault.
///
/// @see MorpheusPublic.kind
#[wasm_bindgen(js_name = MorpheusPublicKind)]
pub struct JsMorpheusPublicKind {
    inner: MorpheusPublicKind,
}

#[wasm_bindgen(js_class = MorpheusPublicKind)]
impl JsMorpheusPublicKind {
    /// Accessor for the kind of identifiers generated by this subtree.
    #[wasm_bindgen(getter)]
    pub fn kind(&self) -> String {
        format!("{:?}", self.inner.path())
    }

    /// Retrieves how many DIDs have already be generated of this kind.
    #[wasm_bindgen(getter)]
    pub fn count(&self) -> Result<u32, JsValue> {
        self.inner.len().map_err_to_js().map(|c| c as u32)
    }

    /// Retrieves the multicipher {@link PublicKey} with the given index in this subtree.
    ///
    /// An error is thrown if that index was not generated yet with {@link MorpheusPrivateKind.key} or {@link MorpheusPrivateKind.did}.
    pub fn key(&self, idx: i32) -> Result<JsMPublicKey, JsValue> {
        let inner = self.inner.key(idx).map_err_to_js()?;
        Ok(JsMPublicKey::from(inner))
    }

    /// Retrieves the {@link Did} with the given index in this subtree.
    ///
    /// An error is thrown if that index was not generated yet with {@link MorpheusPrivateKind.key} or {@link MorpheusPrivateKind.did}.
    pub fn did(&self, idx: i32) -> Result<JsDid, JsValue> {
        let pk = self.inner.key(idx).map_err_to_js()?;
        let inner = Did::from(pk.key_id());
        Ok(JsDid::from(inner))
    }

    /// Finds the multicipher {@link PublicKey} that belongs to the given multicipher {@link KeyId}.
    ///
    /// An error will be thrown if the key has never been used yet in this vault or is not this kind of DID.
    #[wasm_bindgen(js_name = keyById)]
    pub fn key_by_id(&self, id: &JsMKeyId) -> Result<JsMPublicKey, JsValue> {
        let inner = self.inner.key_by_id(id.inner()).map_err_to_js()?;
        Ok(JsMPublicKey::from(inner))
    }
}

impl From<MorpheusPublicKind> for JsMorpheusPublicKind {
    fn from(inner: MorpheusPublicKind) -> Self {
        Self { inner }
    }
}

impl Wraps<MorpheusPublicKind> for JsMorpheusPublicKind {
    fn inner(&self) -> &MorpheusPublicKind {
        &self.inner
    }
}
